module coneneko.controls;
private import hdk.window;
private import hdk.math;

alias Vector.create vector;

interface Control
{
	void input(hdk.window.Window wnd, uint id);
	void draw(hdk.window.Window wnd);
}

template ControlWindow()
{
	private Control[] controlList;
	void addControl(Control a) { controlList ~= a; }
	
	override void writeInput(uint id)
	{
		foreach (Control a; controlList) a.input(this, id);
	}
	
	override void update()
	{
		depthBuffer = false;
		foreach (Control a; controlList) a.draw(this);
		super.update();
	}
	
	void clearControl() { controlList = null; }
}

Matrix toMatrix(float x, float y, float width, float height)
{
	return Matrix.translation(0.5, -0.5, 0)
		* Matrix.scaling(width, height, 1)
		* Matrix.translation(x, y, 0);
}

bit isInnerPoint(float x, float y, float width, float height,
	float mx, float my)
{
	float left = x, right = x + width, top = y, bottom = y - height;
	return left <= mx && mx <= right && bottom <= my && my <= top;
}

class OutputSlider : Control
{
	private float x, y, width, height;
	private float delegate() value;
	
	this(float x, float y, float width, float height, float delegate() value)
	{
		this.x = x;
		this.y = y;
		this.width = width;
		this.height = height;
		this.value = value;
	}
	
	void input(Window wnd, uint id) {}
	
	void draw(Window wnd)
	{
		Vector backColor = vector(1, 1, 1, 1);
		Vector foreColor = vector(1, 0, 0, 1);
		wnd.drawRectangle(backColor, toMatrix(x, y, width, height));
		wnd.drawRectangle(foreColor, toMatrix(x, y, width * value(), height));
	}
}

class InputSlider : Control
{
	private float x, y, width, height;
	private void delegate(float a) setNewValue;
	private float value = 0.0;
	
	this(float x, float y, float width, float height, void delegate(float a) setNewValue)
	{
		this.x = x;
		this.y = y;
		this.width = width;
		this.height = height;
		this.setNewValue = setNewValue;
	}
	
	void input(Window wnd, uint id)
	{
		if (id == 0 && isInnerPoint(wnd.mouseX, wnd.mouseY))
		{
			value = (wnd.mouseY - (y - height)) / height;
			setNewValue(value);
		}
	}
	
	private bit isInnerPoint(float mx, float my)
	{
		return .isInnerPoint(x, y, width, height, mx, my);
	}
	
	void draw(hdk.window.Window wnd)
	{
		Vector backColor = vector(1, 1, 1, 1);
		Vector foreColor = vector(0, 1, 0, 1);
		wnd.drawRectangle(backColor, toMatrix(x, y, width, height));
		float height2 = height * value;
		float y2 = y - height + height2;
		wnd.drawRectangle(foreColor, toMatrix(x, y2, width, height2));
	}
}

class Button : Control
{
	private float x, y, width, height;
	private void delegate() pushButton;
	
	this(float x, float y, float width, float height, void delegate() pushButton)
	{
		this.x = x;
		this.y = y;
		this.width = width;
		this.height = height;
		this.pushButton = pushButton;
	}
	
	void input(Window wnd, uint id)
	{
		if (id == 0 && isInnerPoint(wnd.mouseX, wnd.mouseY))
		{
			pushButton();
		}
	}
	
	private bit isInnerPoint(float mx, float my)
	{
		return .isInnerPoint(x, y, width, height, mx, my);
	}
	
	void draw(Window wnd)
	{
		Vector color = vector(0, 0, 1, 1);
		wnd.drawRectangle(color, toMatrix(x, y, width, height));
	}
}

class CheckButton : Control
{
	private float x, y, width, height;
	private void delegate(bit flag) checkButton;
	private bit flag = false;
	
	this(float x, float y, float width, float height, void delegate(bit flag) checkButton)
	{
		this.x = x;
		this.y = y;
		this.width = width;
		this.height = height;
		this.checkButton = checkButton;
	}
	
	void input(Window wnd, uint id)
	{
		if (id == 0 && isInnerPoint(wnd.mouseX, wnd.mouseY))
		{
			flag = !flag;
			checkButton(flag);
		}
	}
	
	private bit isInnerPoint(float mx, float my)
	{
		return .isInnerPoint(x, y, width, height, mx, my);
	}
	
	void draw(Window wnd)
	{
		Vector falseColor = vector(1, 1, 1, 1);
		Vector trueColor = vector(1, 1, 0, 1);
		wnd.drawRectangle(flag ? trueColor : falseColor, toMatrix(x, y, width, height));
	}
}

class RadioButtonInfo
{
	struct Rectangle { float x, y, width, height; }
	Rectangle[] rectangleList;
	
	void addButton(float x, float y, float width, float height)
	{
		Rectangle rectangle;
		rectangle.x = x;
		rectangle.y = y;
		rectangle.width = width;
		rectangle.height = height;
		rectangleList ~= rectangle;
	}
}

class RadioButton : Control
{
	private RadioButtonInfo info;
	private void delegate(uint index) selectButton;
	private uint index = 0;
	
	this(RadioButtonInfo info, void delegate(uint index) selectButton)
	{
		this.info = info;
		this.selectButton = selectButton;
	}
	
	void input(Window wnd, uint id)
	{
		if (id != 0) return;
		for (uint i = 0; i < info.rectangleList.length; i++)
		{
			RadioButtonInfo.Rectangle a = info.rectangleList[i];
			if (isInnerPoint(a.x, a.y, a.width, a.height, wnd.mouseX, wnd.mouseY))
			{
				if (i == index) return;
				index = i;
				selectButton(index);
			}
		}
	}
	
	void draw(Window wnd)
	{
		Vector falseColor = vector(1, 1, 1, 1);
		Vector trueColor = vector(1, 0, 1, 1);
		foreach (uint i, RadioButtonInfo.Rectangle a; info.rectangleList)
		{
			wnd.drawRectangle(i == index ? trueColor : falseColor,
				toMatrix(a.x, a.y, a.width, a.height));
		}
	}
}
