﻿module hdk.control;
private import hdk.wr;
private import hdk.network;
private import hdk.math;

/*
gluiにあわせる、なければできるだけ短い名前に
Control
Canvas		drawRectangle
ImageBox	drawTexture
StaticText	drawText
Panel		グループ化、不要?
Button
CheckBox
ImageButton
Slider
ProgressBar
CheckBoxGroup
RadioGroup
*/

///
bit isInnerPoint(float x, float y, float width, float height,
	float mx, float my)
{
	float left = x, right = x + width, bottom = y, top = y + height;
	return left <= mx && mx <= right && bottom <= my && my <= top ? true : false;
}

Matrix toMatrix(float x, float y, float width, float height) ///
{
	return Matrix.translation(0.5, 0.5, 0)
		* Matrix.scaling(width, height, 1)
		* Matrix.translation(x, y, 0);
}

Matrix toTextMatrix(float x, float y, float width, float height) ///
{
	return toMatrix(x, y, height * 512.0 / 32.0, height);
}

struct Point
{
	float x, y;
	
	static Point opCall(float x, float y)
	{
		Point result;
		result.x = x;
		result.y = y;
		return result;
	}
}

struct Rectangle
{
	float x, y, width, height;
	
	static Rectangle opCall(float x, float y, float width, float height)
	{
		Rectangle result;
		result.x = x;
		result.y = y;
		result.width = width;
		result.height = height;
		return result;
	}
	
	bit inner(Point p)
	{
		return .isInnerPoint(x, y, width, height, p.x, p.y);
	}
	
	Matrix toMatrix()
	{
		return .toMatrix(x, y, width, height);
	}
	
	Matrix toTextMatrix()
	{
		return .toTextMatrix(x, y, width, height);
	}
	
	alias x left;
	alias y bottom;
	float right() { return x + width; }
	float top() { return y + height; }
}

class Control : Node
{
	protected static Network network;
	protected static WindowResource wr;
	private bit _visible = true;
	
	float z = 0.0; ///
	
	this(Control parent) ///
	{
		network ~= this;
		network.link(parent, this);
		z = parent.z + 1.0;
	}
	
	protected this(WindowResource wr)
	{
		if (!network) network = new Network();
		network ~= this;
		this.wr = wr;
	}
	
	~this()
	{
		cutLink();
	}
	
	void cutLink() ///
	{
		Node p = parent;
		if (p) network.cut(p, this);
	}
	
	protected Control parent()
	{
		return cast(Control)network.parentOf(this);
	}
	
	bit inner(Point p) { return false; }
	void input(uint id) {}
	void draw() {}
	
	static protected Point mousePosition()
	{
		return Point(wr.mouseX, wr.mouseY);
	}
	
	protected Control[] children()
	{
		return cast(Control[])network.childrenOf(this);
	}
	
	bit visible()
	{
		return _visible;
	}
	
	void visible(bit a)
	{
		_visible = a;
		foreach (inout Control c; children)
		{
			c.visible = a;
		}
	}
}

class RootControl : Control
{
	this(WindowResource wr)
	{
		super(wr);
	}
	
	void clearOtherTree()
	{
		Node thisRoot = network.rootOf(this);
		Node[] otherNodeList;
		foreach (Node a; network)
		{
			Node r = network.rootOf(a);
			if (r != thisRoot) otherNodeList ~= r;
		}
		foreach (Node a; otherNodeList) network.clear(a);
	}
	
	private Control[] sortedControlList;
	
	override void input(uint id)
	{
		for (int i = sortedControlList.length - 1; i >= 0; i--)
		{
			if (sortedControlList[i].visible
				&& sortedControlList[i].inner(mousePosition))
			{
				sortedControlList[i].input(id);
				return;
			}
		}
	}
	
	override void draw()
	{
		zsort();
		foreach (Control c; sortedControlList)
		{
			if (c.visible) c.draw();
		}
	}
	
	private void zsort()
	{
		Control[] controlList;
		foreach (Node n; network)
		{
			if (n != this) controlList ~= cast(Control)n;
		}
		
		float[] zList;
		foreach (Control c; controlList) zList ~= c.z;
		
		zList.sort;
		
		Control[] newSortedControlList;
		foreach (float z; zList)
		{
			foreach (inout Control c; controlList)
			{
				if (c && c.z == z)
				{
					newSortedControlList ~= c;
					c = null;
					break;
				}
			}
		}
		
		sortedControlList = newSortedControlList;
	}
	
	void setWindowResource(WindowResource a)
	{
		wr = a;
	}
}

class Billboard : Control
{
	bit depthTest = false, depthWriting = false,
		cullingEnable = false, cullingFront = false;
	Matrix modelView, projection;
	Vector color;
	
	this(Control parent)
	{
		super(parent);
		projection = modelView = Matrix.identity;
		color = Vector(1, 1, 1, 1);
	}
	
	protected void setRenderState()
	{
		wr.depthTestWriting(depthTest, depthWriting);
		wr.cullingEnableFront(cullingEnable, cullingFront);
		wr.modelView = modelView;
		wr.projection = projection;
	}
}

class Canvas : Billboard
{
	this(Control parent)
	{
		super(parent);
	}
	
	override void draw()
	{
		setRenderState();
		wr.drawRectangle(color);
	}
}

class ImageBox : Billboard
{
	dchar[] textureFileName;
	
	this(Control parent)
	{
		super(parent);
	}
	
	override void draw()
	{
		if (!textureFileName) return;
		setRenderState();
		wr.drawTexture(textureFileName, color);
	}
}

class StaticText : Billboard
{
	dchar[] text;
	
	this(Control parent)
	{
		super(parent);
	}
	
	override void draw()
	{
		if (!text) return;
		setRenderState();
		wr.drawText(text, color);
	}
}

class Button : Control
{
	private final Canvas canvas;
	private final StaticText staticText;
	private Rectangle _rectangle;
	
	dchar[] text;
	Vector foreColor, backColor;
	void delegate() pushedEvent;
	
	this(Control parent)
	{
		super(parent);
		canvas = new Canvas(this);
		staticText = new StaticText(canvas);
	}
	
	override bit inner(Point p)
	{
		return _rectangle.inner(p);
	}
	
	override void input(uint id)
	{
		if (id == 0 && inner(mousePosition) && pushedEvent) pushedEvent();
	}
	
	override void draw()
	{
		canvas.color = backColor;
		canvas.color.a = backAlpha;
		staticText.text = text;
		staticText.color = foreColor;
	}
	
	private float backAlpha()
	{
		const float OFF_ALPHA = 0.4, FLOATING_ALPHA = 0.7, ON_ALPHA = 0.9;
		return inner(mousePosition) ? FLOATING_ALPHA : OFF_ALPHA;
	}
	
	void rectangle(Rectangle r)
	{
		_rectangle = r;
		canvas.modelView = r.toMatrix();
		staticText.modelView = r.toTextMatrix();
	}
}

class ProgressBar : Control
{
	private final Canvas foreCanvas, backCanvas;
	
	float value = 0.0; ///
	Rectangle rectangle;
	Vector foreColor, backColor;
	
	this(Control parent)
	{
		super(parent);
		backCanvas = new Canvas(this);
		foreCanvas = new Canvas(backCanvas);
	}
	
	override void draw()
	{
		backCanvas.modelView = rectangle.toMatrix();
		backCanvas.color = backColor;
		foreCanvas.modelView = foreRectangle.toMatrix();
		foreCanvas.color = foreColor;
	}
	
	private Rectangle foreRectangle()
	{
		Rectangle result = rectangle;
		result.width *= value;
		return result;
	}
}

class Slider : Control
{
	private final Canvas foreCanvas, backCanvas;
	
	Rectangle rectangle;
	Vector foreColor, backColor;
	void delegate(float value) changedEvent;
	
	private float _value = 0.0;
	private bit dragging = false;
	
	this(Control parent)
	{
		super(parent);
		backCanvas = new Canvas(this);
		foreCanvas = new Canvas(backCanvas);
	}
	
	void value(float a)
	{
		_value = clamp(a, 0.0, 1.0);
		if (changedEvent) changedEvent(_value);
	}
	
	float value()
	{
		return _value;
	}
	
	override bit inner(Point p)
	{
		return rectangle.inner(p);
	}
	
	override void input(uint id)
	{
		if (id == 0 && inner(mousePosition)) dragging = true;
		else if (id == 1 && dragging)
		{
			value = computeValue();
			dragging = false;
		}
	}
	
	override void draw()
	{
		if (dragging && !inner(mousePosition))
		{
			value = computeValue();
			dragging = false;
		}
		
		backCanvas.modelView = rectangle.toMatrix();
		backCanvas.color = backColor;
		foreCanvas.modelView = foreRectangle.toMatrix();
		foreCanvas.color = foreColor;
	}
	
	private float computeValue()
	{
		return clamp((wr.mouseY - rectangle.bottom) / rectangle.height, 0.0, 1.0);
	}
	
	private Rectangle foreRectangle()
	{
		Rectangle result = rectangle;
		result.height *= dragging ? computeValue() : _value;
		return result;
	}
}

class CheckBox : Control
{
	private final Canvas canvas;
	private final StaticText staticText;
	
	dchar[] text;
	Rectangle rectangle;
	Vector foreColor, backColor;
	void delegate(bit value) checkedEvent;
	bit value = false;;
	
	this(Control parent)
	{
		super(parent);
		canvas = new Canvas(this);
		staticText = new StaticText(canvas);
	}
	
	override bit inner(Point p)
	{
		return rectangle.inner(p);
	}
	
	override void input(uint id)
	{
		if (id == 0 && inner(mousePosition))
		{
			value = !value ? true : false;
			if (checkedEvent) checkedEvent(value);
		}
	}
	
	override void draw()
	{
		canvas.modelView = rectangle.toMatrix();
		canvas.color = backColor;
		canvas.color.a = backAlpha;
		staticText.text = text;
		staticText.modelView = rectangle.toTextMatrix();
		staticText.color = foreColor;
	}
	
	private float backAlpha()
	{
		const float[] alphaList = [ 0.5, 0.6, 0.9, 1.0 ];
		return alphaList[(value ? 2 : 0) + (inner(mousePosition) ? 1 : 0)];
	}
}

class CheckBoxGroup : Control
{
	private final Canvas borderCanvas;
	private CheckBox[] cbList;
	
	Vector foreColor, backColor, borderColor;
	Rectangle rectangle;
	
	this(Control parent)
	{
		super(parent);
		borderCanvas = new Canvas(this);
	}
	
	void length(uint a)
	{
		cbList.length = a;
		foreach (inout CheckBox cb; cbList) if (!cb) cb = new CheckBox(borderCanvas);
	}
	
	uint length()
	{
		return cbList.length;
	}
	
	CheckBox opIndex(uint index)
	{
		return cbList[index];
	}
	
	override void draw()
	{
		borderCanvas.modelView = rectangle.toMatrix();
		borderCanvas.color = borderColor;
		
		Rectangle[] rectList = splitRectangle(rectangle, cbList.length);
		addMargin(rectList, rectangle.height * 0.04);
		
		for (int i = 0; i < cbList.length; i++)
		{
			cbList[i].rectangle = rectList[i];
			cbList[i].foreColor = foreColor;
			cbList[i].backColor = backColor;
		}
	}
	
	private static Rectangle[] splitRectangle(Rectangle base, uint itemLength)
	{
		Rectangle[] result;
		for (int i = 0; i < itemLength; i++)
		{
			float itemWidth = base.width / itemLength;
			float itemX = base.x + itemWidth * i;
			result ~= Rectangle(itemX, base.y, itemWidth, base.height);
		}
		return result;
	}
	
	private static void addMargin(inout Rectangle[] list, float margin)
	{
		foreach (inout Rectangle a; list)
		{
			a.x += margin;
			a.y += margin;
			a.width -= margin * 2;
			a.height -= margin * 2;
		}
	}
}

class RadioGroup : CheckBoxGroup
{
	void delegate(uint index) selectedEvent;
	uint value = 0;
	
	this(Control parent)
	{
		super(parent);
	}
	
	override void draw()
	{
		uint cs = countSelection();
		if (cs == 0)
		{
			if (cbList.length >= value) cbList[value].value = true;
			value = currentSelection();
			selectedEvent(value);
		}
		else if (cs >= 2)
		{
			cbList[value].value = false;
			value = currentSelection();
			selectedEvent(value);
		}
		
		super.draw();
	}
	
	private uint countSelection()
	{
		uint result = 0;
		foreach (CheckBox a; cbList) if (a.value) result++;
		return result;
	}
	
	private uint currentSelection()
	{
		foreach (int i, CheckBox a; cbList) if (a.value) return i;
		throw new Error("currentSelection");
	}
}

// TODO coneneko側に、SaveDataButton?
class ImageButton : Button
{
	private final ImageBox imageBox;
	
	float imageWidth, textHeight;
	
	this(Control parent)
	{
		super(parent);
		imageBox = new ImageBox(this);
	}
	
	override void draw()
	{
		Rectangle textRectangle = _rectangle;
		textRectangle.x += imageWidth;
		textRectangle.y += (textRectangle.height - textHeight) * 0.5;
		textRectangle.height = textHeight;
		staticText.modelView = textRectangle.toTextMatrix();
		
		Rectangle imageRectangle = _rectangle;
		imageRectangle.width = imageWidth;
		imageBox.modelView = imageRectangle.toMatrix();
		
		super.draw();
	}
	
	void textureFileName(dchar[] a)
	{
		imageBox.textureFileName = a;
	}
}
