﻿module hdk.oglshader;
private import sdl;
private import hdk.error;
private import hdk.openglarb;
private import std.string;
private import std.random;
private import std.file;
private import std.utf;

interface OglShaderMap
{
	bit hasShader(dchar[] key);
	void addShader(dchar[] key, char[] vertexSource, char[] fragmentSource);
	void setShaderFloat(dchar[] key, char[] shaderKey, float a);
	void setShaderInt(dchar[] key, char[] shaderKey, int a);
	void setShaderVector(dchar[] key, char[] shaderKey, float a, float b, float c, float d);
	void setShaderMatrixArray(dchar[] key, char[] shaderKey, float* p, uint matrixLength);
	uint getShaderAttributeLocation(dchar[] key, char[] shaderKey);
	void attachShader(dchar[] key);
	void detachShader();
	void clearShaderMap();
}

private class OglShader
{
	private import hdk.error;
	mixin OglError;
	
	private uint shader;
	private bit lock;
	
	this(char[] vertexSource, char[] fragmentSource)
	{
		shader = createShader(vertexSource, fragmentSource);
	}
	
	~this()
	{
		glDeleteObjectARB(shader);
	}
	
	void bind() { glUseProgramObjectARB(shader); }
	
	// 使わない値には代入できないようなので注意
	static assert(GLint.max == int.max);
	private static int getUniformLocation(uint id, char[] key)
	{
		int result = glGetUniformLocationARB(id, toStringz(key));
		if (-1 == result) throw new Error("UniformLocation: " ~ key);
		return result;
	}
	
	void setFloat(char[] key, float a)
	{
		bind();
		glUniform1fARB(getUniformLocation(shader, key), a);
	}
	
	float getFloat(char[] key)
	{
		static assert(float.sizeof == GLfloat.sizeof);
		float result;
		bind();
		glGetUniformfvARB(shader, getUniformLocation(shader, key), &result);
		return result;
	}
	
	void setInt(char[] key, int a)
	{
		bind();
		glUniform1iARB(getUniformLocation(shader, key), a);
	}
	
	int getInt(char[] key)
	{
		static assert(int.sizeof == GLint.sizeof);
		int result;
		bind();
		glGetUniformivARB(shader, getUniformLocation(shader, key), &result);
		return result;
	}
	
	void setVector(char[] key, float a, float b, float c, float d)
	{
		bind();
		glUniform4fARB(getUniformLocation(shader, key), a, b, c, d);
	}
	
	void getVector(char[] key, out float x, out float y, out float z, out float w)
	{
		float[] result = new float[4];
		bind();
		glGetUniformfvARB(shader, getUniformLocation(shader, key), result.ptr);
		x = result[0];
		y = result[1];
		z = result[2];
		w = result[3];
	}
	
	void setMatrixArray(char[] key, float* p, uint matrixLength) // Matrix[]
	{
		bind();
		glUniformMatrix4fvARB(
			getUniformLocation(shader, key), matrixLength, GL_FALSE, p);
	}
	
	float[] getMatrixArray(char[] key, uint matrixLength)
	{
		float[] result = new float[4 * 4 * matrixLength];
		bind(); //?
		glGetUniformfvARB(shader, getUniformLocation(shader, key), result.ptr);
		return result;
	}
	
	uint getAttributeLocation(char[] key)
	{
		bind();
		int result = glGetAttribLocationARB(shader, toStringz(key));
		if (-1 == result) error(__LINE__, "getAttributeLocation: " ~ key);
		return result;
	}
	
	protected int createShader(char[] vertexSource, char[] fragmentSource)
	{
		const char[] errorMessage = "WindowResourceLayer1.createShader";
		GLhandleARB result;
		GLhandleARB vsObj, fsObj;
		try
		{
			result = glCreateProgramObjectARB();
			if (!result) error(__LINE__, errorMessage);
			
			vsObj = compileShader(GL_VERTEX_SHADER_ARB, vertexSource);
			fsObj = compileShader(GL_FRAGMENT_SHADER_ARB, fragmentSource);
			
			glAttachObjectARB(result, vsObj);
			glAttachObjectARB(result, fsObj);
			
			glLinkProgramARB(result);
			GLint resultL;
			glGetObjectParameterivARB(result, GL_OBJECT_LINK_STATUS_ARB, &resultL);
			if (!resultL) error(__LINE__, "shader link error");
		}
		catch (Exception e) { throw e; }
		finally
		{
			if (fsObj) glDeleteObjectARB(fsObj);
			if (vsObj) glDeleteObjectARB(vsObj);
		}
		return result;
	}
	
	static assert(int.max == GLhandleARB.max);
	protected int compileShader(GLenum shaderType, char[] source)
	{
		const char[] errorMessage = "WindowResourceLayer1.compileShader";
		GLhandleARB result = glCreateShaderObjectARB(shaderType);
		if (!result) error(__LINE__, errorMessage);
		
		char* sourcez = toStringz(source);
		int sourceLength = source.length;
		glShaderSourceARB(result, 1, &sourcez, &sourceLength);

		int resultC;
		glCompileShaderARB(result);
		glGetObjectParameterivARB(result, GL_OBJECT_COMPILE_STATUS_ARB, &resultC);
		if (!resultC)
		{
			int infoLogLength;
			glGetObjectParameterivARB(result, GL_OBJECT_INFO_LOG_LENGTH_ARB, &infoLogLength);
			assert(0 < infoLogLength);
			char[] infoLog = new char[infoLogLength];
			GLsizei r;
			glGetInfoLogARB(result, infoLog.length, &r, infoLog.ptr);
			assert(infoLog.length - 1 == r);
			error(__LINE__, "shader compile error: " ~ infoLog);
		}
		return result;
	}
}

template OglShaderMap_()
{
	private import hdk.error;
	mixin SdlError;
	
	private static bit initialized = false;
	
	private OglShader[dchar[]] map;
	private bit lock = false;
	
	this()
	{
		if (initialized) error(__LINE__, "既に初期化されている");
		initialized = true;
	}
	
	~this()
	{
		clearShaderMap();
		initialized = false;
	}
	
	void clearShaderMap()
	{
		if (lock) detachShader();
		foreach (OglShader a; map) delete a;
		map = null;
	}
	
	bit hasShader(dchar[] key)
	{
		throwIfLockError();
		return key in map ? true : false;
	}
	
	void addShader(dchar[] key, char[] vertexSource, char[] fragmentSource)
	{
		throwIfLockError();
		if (hasShader(key)) error(__LINE__, "addShader");
		map[key] = new OglShader(vertexSource, fragmentSource);
	}
	
	void setShaderFloat(dchar[] key, char[] shaderKey, float a)
	{
		throwIfLockError();
		throwIfNotFoundError(key);
		map[key].setFloat(shaderKey, a);
	}
	
	void setShaderInt(dchar[] key, char[] shaderKey, int a)
	{
		throwIfLockError();
		throwIfNotFoundError(key);
		map[key].setInt(shaderKey, a);
	}
	
	void setShaderVector(dchar[] key, char[] shaderKey, float a, float b, float c, float d)
	{
		throwIfLockError();
		throwIfNotFoundError(key);
		map[key].setVector(shaderKey, a, b, c, d);
	}
	
	void setShaderMatrixArray(dchar[] key, char[] shaderKey, float* p, uint matrixLength)
	{
		throwIfLockError();
		throwIfNotFoundError(key);
		map[key].setMatrixArray(shaderKey, p, matrixLength);
	}
	
	uint getShaderAttributeLocation(dchar[] key, char[] shaderKey)
	{
		throwIfLockError();
		throwIfNotFoundError(key);
		return map[key].getAttributeLocation(shaderKey);
	}
	
	void attachShader(dchar[] key)
	{
		throwIfLockError();
		throwIfNotFoundError(key);
		map[key].bind();
		lock = true;
	}
	
	void detachShader()
	{
		if (!lock) error(__LINE__, "detach");
		lock = false;
	}
	
	private void throwIfLockError()
	{
		if (lock) error(__LINE__, "lock is true");
	}
	
	private void throwIfNotFoundError(dchar[] key)
	{
		if (hasShader(key)) return;
		try
		{
			char[] vertexSource = cast(char[])read(toUTF8(key ~ ".vert"));
			char[] fragmentSource = cast(char[])read(toUTF8(key ~ ".frag"));
			addShader(key, vertexSource, fragmentSource);
		}
		catch (Error e)
		{
			error(__LINE__, format("%s not found", key));
		}
	}
}

auto class OglShaderMapTest : OglShaderMap
{
	private import std.cstream;
	mixin OglShaderMap_;
	
	static void execute()
	{
		try
		{
			initialize();
			initializeArb();
			testOglShader();
			testOglShaderMap();
		}
		catch (Exception e)
		{
			derr.writeLine("Exception:" ~ e.toString());
		}
		finally
		{
			SDL_Quit();
		}
	}
	
	private static void initialize()
	{
		uint WIDTH = 1, HEIGHT = 1;
		SDL_Init(SDL_INIT_VIDEO);
		SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 8);
		SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 8);
		SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, 8);
		SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 16);
		SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
		SDL_SetVideoMode(WIDTH, HEIGHT, 24, SDL_OPENGL);
	}
	
	static void testOglShader()
	{
		const char[] vs = "
		attribute vec2 matrixIndex;
		attribute float weight;
		uniform float f;
		uniform int i;
		uniform vec4 v;
		uniform mat4 ma[2];
		void main()
		{
			matrixIndex, weight, f, i, v, ma;
			gl_Position = gl_Vertex;
			gl_FrontColor = gl_Color;
		}
		";
		const char[] fs = "
		uniform sampler2D texture0, texture1;
		void main()
		{
			texture0, texture1;
			gl_FragColor = gl_Color;
		}
		";
		auto OglShader shader = new OglShader(vs, fs);
		assert(2 == shader.getAttributeLocation("matrixIndex")); // 適当
		assert(3 == shader.getAttributeLocation("weight")); // 適当
		try
		{
			shader.getAttributeLocation("hoge");
		}
		catch (Exception e)
		{
			assert("OglShader getAttributeLocation: hoge 117" == e.toString());
		}
		shader.setFloat("f", 2.0);
		assert(2.0 == shader.getFloat("f"));
		shader.setInt("i", 3);
		assert(3 == shader.getInt("i"));
		shader.setVector("v", 4, 5, 6, 7);
		struct vec { float x, y, z, w; }
		vec vr;
		float a;
		shader.getVector("v", vr.x, vr.y, vr.z, vr.w);
		assert(4 == vr.x);
		assert(5 == vr.y);
		assert(6 == vr.z);
		assert(7 == vr.w);
		float[] testMatrix = new float[4 * 4 * 2];
		foreach (inout float a; testMatrix) a = cast(float)rand();
		shader.setMatrixArray("ma", testMatrix.ptr, 2);
		assert(testMatrix == shader.getMatrixArray("ma", 2));
		
		auto OglShader shader2 = new OglShader(vs, fs);
		shader2.setFloat("f", 3.0);
		assert(3.0 == shader2.getFloat("f"));
		assert(2.0 == shader.getFloat("f"));
		
		shader.setInt("texture0", 0); // 多分マルチテクスチャのインデックス
		shader.setInt("texture1", 1);
		//assert(0 == shader.getInt("texture0"));
		//assert(1 == shader.getInt("texture1"));
	}
	
	static void testOglShaderMap()
	{
		const char[] vs = "
		attribute vec2 matrixIndex;
		attribute float weight;
		uniform float f;
		uniform int i;
		uniform vec4 v;
		uniform mat4 ma[2];
		void main()
		{
			matrixIndex, weight, f, i, v, ma;
			gl_Position = gl_Vertex;
			gl_FrontColor = gl_Color;
		}
		";
		const char[] fs = "
		uniform sampler2D texture0, texture1;
		void main()
		{
			texture0, texture1;
			gl_FragColor = gl_Color;
		}
		";
		auto OglShaderMapTest test = new OglShaderMapTest();
		with (test)
		{
			assert(!hasShader("hoge"));
			addShader("hoge", vs, fs);
			assert(hasShader("hoge"));
			attachShader("hoge");
			try
			{
				hasShader("hoge");
			}
			catch (Exception e)
			{
				assert("OglShaderMapTest lock is true 273" == e.toString());
			}
			detachShader();
		}
	}
}
