﻿module hdk.ogltexture;
private import sdl;
private import opengl;
private import std.random;

interface OglTextureMap
{
	bit hasTexture(dchar[] key);
	void addTexture(dchar[] key, uint[] rgba, uint width, uint height);
	uint getTextureWidth(dchar[] key);
	uint getTextureHeight(dchar[] key);
	uint[] getTextureRgbaPixels(dchar[] key);
	void attachTexture(dchar[] key0, dchar[] key1, dchar[] key2, dchar[] key3);
	void detachTexture();
	void clearTextureMap();
	uint totalTextureSize();
}

private class OglTexture
{
	private import hdk.error;
	mixin SdlError;
	
	private GLuint texture;
	
	this(uint[] rgba, uint width, uint height)
	{
		if (rgba.length != width * height) error(__LINE__, "テクスチャのサイズが変");
		throwIfTextureSizeError(width, height);
		glGenTextures(1, &texture);
		glBindTexture(GL_TEXTURE_2D, texture);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, width, height, 0,
			GL_RGBA, GL_UNSIGNED_BYTE, rgba.ptr);
	}
	
	~this()
	{
		glDeleteTextures(1, &texture);
	}
	
	uint getWidth()
	{
		GLint result;
		bind();
		glGetTexLevelParameteriv(GL_TEXTURE_2D, 0, GL_TEXTURE_WIDTH, &result);
		return result;
	}
	
	uint getHeight()
	{
		GLint result;
		bind();
		glGetTexLevelParameteriv(GL_TEXTURE_2D, 0, GL_TEXTURE_HEIGHT, &result);
		return result;
	}
	
	uint[] getRgbaPixels()
	{
		uint[] result = new uint[getWidth() * getHeight()];
		bind();
		glGetTexImage(GL_TEXTURE_2D, 0, GL_RGBA, GL_UNSIGNED_BYTE, result.ptr);
		return result;
	}
	
	void bind() { glBindTexture(GL_TEXTURE_2D, texture); }
	
	void throwIfTextureSizeError(uint width, uint height)
	{
		throwIfTextureSizeError(width);
		throwIfTextureSizeError(height);
	}
	
	void throwIfTextureSizeError(uint a)
	{
		while (a != 1)
		{
			if (a % 2 != 0) error(__LINE__, "throwIfTextureSizeError");
			a /= 2;
		}
	}
}

template OglTextureMap_()
{
	private import hdk.openglarb;
	private import hdk.error;
	mixin SdlError;
	
	private static bit initialized = false;
	
	private const uint ACTIVE_TEXTURE_MAX = 4; // 0, 1, 2, 3まで使う
	private OglTexture[dchar[]] textureMap;
	private bit lock = false;
	
	this()
	{
		if (initialized) error(__LINE__, "既に初期化されてる");
		initializeArb();
		initialized = true;
	}
	
	~this()
	{
		clearTextureMap();
		initialized = false;
	}
	
	void clearTextureMap()
	{
		if (lock) detachTexture();
		foreach (OglTexture a; textureMap.values) delete a;
		textureMap = null;
		totalSize = 0;
	}
	
	bit hasTexture(dchar[] key)
	{
		throwIfLockError();
		return key in textureMap ? true : false;
	}
	
	void addTexture(dchar[] key, uint[] rgba, uint width, uint height)
	{
		throwIfTextureSizeError(rgba, width, height);
		throwIfLockError();
		if (hasTexture(key)) return;
		textureMap[key] = new OglTexture(rgba, width, height);
		totalSize += uint.sizeof * width * height;
	}
	
	uint getTextureWidth(dchar[] key)
	{
		throwIfLockError();
		throwIfNotFoundError(key);
		return textureMap[key].getWidth();
	}
	
	uint getTextureHeight(dchar[] key)
	{
		throwIfLockError();
		throwIfNotFoundError(key);
		return textureMap[key].getHeight();
	}
	
	uint[] getTextureRgbaPixels(dchar[] key)
	{
		throwIfLockError();
		throwIfNotFoundError(key);
		return textureMap[key].getRgbaPixels();
	}
	
	void attachTexture(dchar[] key0, dchar[] key1 = null, dchar[] key2 = null, dchar[] key3 = null)
	{
		throwIfLockError();
		
		dchar[][] keys;
		keys ~= key0;
		keys ~= key1;
		keys ~= key2;
		keys ~= key3;
		assert(4 == keys.length);
		
		foreach (dchar[] a; keys)
		{
			if (a) throwIfNotFoundError(a);
		}
		
		for (int i = 0; i < ACTIVE_TEXTURE_MAX; i++)
		{
			activeTexture(i);
			if (!keys[i]) continue;
			glEnable(GL_TEXTURE_2D);
			textureMap[keys[i]].bind();
		}
		lock = true;
	}
	
	void detachTexture()
	{
		for (int i = 0; i < ACTIVE_TEXTURE_MAX; i++)
		{
			activeTexture(i);
			glDisable(GL_TEXTURE_2D);
		}
		lock = false;
	}
	
	protected void activeTexture(uint index)
	{
		if (ACTIVE_TEXTURE_MAX <= index) error(__LINE__, "activeTexture");
		glActiveTextureARB(GL_TEXTURE0_ARB + index);
	}
	
	private void throwIfNotFoundError(dchar[] key)
	{
		if (!hasTexture(key)) error(__LINE__, format("not found %s texture", key));
	}
	
	private void throwIfLockError()
	{
		if (lock) error(__LINE__, "lock is true");
	}
	
	private uint totalSize = 0;
	uint totalTextureSize() { return totalSize; }
	
	private void throwIfTextureSizeError(uint[] rgba, uint width, uint height)
	{
		bit rgbaLengthIsCorrect = rgba.length == width * height ? true : false;
		
		bit widthIsCorrect = false, heightIsCorrect = false;
		const uint[] list = [ 1, 2, 4, 8, 16, 32, 64, 128, 256, 512, 1024, 2048 ];
		foreach (uint a; list)
		{
			if (a == width) widthIsCorrect = true;
			if (a == height) heightIsCorrect = true;
		}
		
		if (rgbaLengthIsCorrect && widthIsCorrect && heightIsCorrect) return;
		
		error(__LINE__, format("rgba.length=%d, width=%d, height=%d", rgba.length, width, height));
	}
}

auto class OglTextureMapTestBase
{
	this()
	{
		const uint WIDTH = 1, HEIGHT = 1;
		SDL_Init(SDL_INIT_VIDEO);
		SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 8);
		SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 8);
		SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, 8);
		SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 16);
		SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
		SDL_SetVideoMode(WIDTH, HEIGHT, 24, SDL_OPENGL);
	}
	
	~this() { SDL_Quit(); }
}

auto class OglTextureMapTest : OglTextureMapTestBase, OglTextureMap
{
	mixin OglTextureMap_;
	
	static void execute()
	{
		auto OglTextureMapTest test = new OglTextureMapTest();
		with (test)
		{
			const uint width = 256, height = 128;
			uint[] testData = new uint[width * height];
			foreach (inout uint a; testData) a = rand();
			
			assert(!hasTexture("hoge"));
			addTexture("hoge", testData, width, height);
			assert(hasTexture("hoge"));
			assert(256 == getTextureWidth("hoge"));
			assert(128 == getTextureHeight("hoge"));
			assert(testData == getTextureRgbaPixels("hoge"));
			
			const uint width2 = 128, height2 = 64;
			uint[] testData2 = new uint[width2 * height2];
			foreach (inout uint a; testData2) a = rand();
			
			assert(!hasTexture("hoge2"));
			addTexture("hoge2", testData2, width2, height2);
			assert(hasTexture("hoge2"));
			assert(128 == getTextureWidth("hoge2"));
			assert(64 == getTextureHeight("hoge2"));
			assert(testData2 == getTextureRgbaPixels("hoge2"));
			
			assert(256 == getTextureWidth("hoge"));
			assert(64 == getTextureHeight("hoge2"));
			assert(testData != testData2);
			assert(testData == getTextureRgbaPixels("hoge"));
			assert(testData2 == getTextureRgbaPixels("hoge2"));
			
			static assert(true == GL_TRUE);
			static assert(false == GL_FALSE);
			bit testEnabledTexture2(bit i0, bit i1, bit i2, bit i3)
			{
				activeTexture(0);
				if (glIsEnabled(GL_TEXTURE_2D) != i0) return false;
				activeTexture(1);
				if (glIsEnabled(GL_TEXTURE_2D) != i1) return false;
				activeTexture(2);
				if (glIsEnabled(GL_TEXTURE_2D) != i2) return false;
				activeTexture(3);
				if (glIsEnabled(GL_TEXTURE_2D) != i3) return false;
				return true;
			}
			
			assert(testEnabledTexture2(false, false, false, false));
			attachTexture("hoge");
			assert(testEnabledTexture2(true, false, false, false));
			detachTexture();
			assert(testEnabledTexture2(false, false, false, false));
			attachTexture("hoge2", "hoge");
			assert(testEnabledTexture2(true, true, false, false));
			
			Exception stock;
			try getTextureWidth("hoge2");
			catch (Exception e) { stock = e; }
			finally { assert("OglTextureMapTest lock is true 202" == stock.toString()); }
			
			detachTexture();
			assert(testEnabledTexture2(false, false, false, false));
		}
	}
}
