﻿module hdk.sdlinput;
private import std.openrj;
private import hdk.sdlbase;

template SdlInput()
{
	private import hdk.error;
	private import std.file;
	mixin SdlError;
	
	private static InputEventTranslator iet;
	
	private SDL_Joystick* joystick;
	private float _mouseX = 0.0, _mouseY = 0.0;
	private const char[] INPUT_RJ_FILE_NAME = "input.rj";
	
	this()
	{
		if (iet || !SDL_GetVideoSurface()) error(__LINE__);
		iet = new InputEventTranslator(cast(char[])read(INPUT_RJ_FILE_NAME));
		initializeJoystick();
	}
	
	~this()
	{
		if (joystick) SDL_JoystickClose(joystick);
		SDL_QuitSubSystem(SDL_INIT_JOYSTICK);
		delete iet;
	}
	
	private void initializeJoystick()
	{
		SDL_InitSubSystem(SDL_INIT_JOYSTICK);
		if (SDL_NumJoysticks() == 0) return;
		joystick = SDL_JoystickOpen(0);
		assert(SDL_ENABLE == SDL_JoystickEventState(SDL_QUERY));
	}
	
	int pollEvent(void* sdlEvent) { return pollEvent2(cast(SDL_Event*)sdlEvent); }
	
	private int pollEvent2(SDL_Event* event)
	{
		while (true)
		{
			if (0 == SDL_PollEvent(event)) return 0;
			
			if (event.type == SDL_MOUSEMOTION)
			{
				_mouseX = cast(float)event.motion.x / cast(float)winWidth * 2.0 - 1.0;
				_mouseY = -(cast(float)event.motion.y / cast(float)winHeight * 2.0 - 1.0);
				continue;
			}
			else if (isNotInput(*event)) return 1;
			
			uint input = iet.execute(getInputEvent(*event));
			if (input != uint.max)
			{
				inputBuffer ~= input;
				if (inputBuffer.length > INPUT_BUFFER_MAX) inputBuffer.length = 0;
				
				if (_inputHandler) _inputHandler(input);
			}
		}
		error(__LINE__);
		return 0;
	}
	
	private uint winWidth() { return SDL_GetVideoSurface().w; }
	private uint winHeight() { return SDL_GetVideoSurface().h; }
	
	float mouseX() { return _mouseX; }
	float mouseY() { return _mouseY; }
	
	private static bit isNotInput(SDL_Event event)
	{
		return !isMouseType(event)
			&& !isKeyboardType(event)
			&& !isJoystickType(event) ? true : false;
	}
	
	private static bit isMouseType(SDL_Event event)
	{
		return event.type == SDL_MOUSEBUTTONUP
			|| event.type == SDL_MOUSEBUTTONDOWN ? true : false;
	}
	
	private static bit isKeyboardType(SDL_Event event)
	{
		return event.type == SDL_KEYUP
			|| event.type == SDL_KEYDOWN ? true : false;
	}
	
	private static bit isJoystickType(SDL_Event event)
	{
		return event.type == SDL_JOYAXISMOTION
			|| event.type == SDL_JOYBUTTONDOWN
			|| event.type == SDL_JOYBUTTONUP ? true : false;
	}
	
	private InputEvent getInputEvent(SDL_Event event)
	{
		InputEvent result;
		result.deviceType = isMouseType(event) ? InputEvent.DeviceType.MOUSE
			: isKeyboardType(event) ? InputEvent.DeviceType.KEYBOARD
			: InputEvent.DeviceType.JOYSTICK;
		result.isOn = event.type == SDL_MOUSEBUTTONDOWN
			|| event.type == SDL_KEYDOWN
			|| isJoystickButtonOn(event) ? true : false;
		result.code = result.isMouse ? event.button.button
			: result.isKeyboard ? event.key.keysym.sym
			: getJoystickCode(event);
		return result;
	}
	
	enum JoystickCrossState { FREE, UP, DOWN, RIGHT, LEFT }
	alias JoystickCrossState Jcs;
	private static Jcs prejcs = Jcs.FREE, jcs = Jcs.FREE;
	
	private static bit isJoystickButtonOn(SDL_Event event)
	{
		if (event.type == SDL_JOYAXISMOTION)
		{
			with (event.jaxis)
			{
				if (axis != 0 && axis != 1) return false;
				prejcs = jcs;
				jcs = Jcs.FREE;
				if (axis == 1 && value < short.min / 2) jcs = Jcs.UP;
				else if (axis == 1 && short.max / 2 < value) jcs = Jcs.DOWN;
				else if (axis == 0 && short.max / 2 < value) jcs = Jcs.RIGHT;
				else if (axis == 0 && value < short.min / 2) jcs = Jcs.LEFT;
			}
			return jcs != Jcs.FREE ? true : false;
		}
		else if (event.type == SDL_JOYBUTTONDOWN) return true;
		else if (event.type == SDL_JOYBUTTONUP) return false;
		return false;
	}
	
	private uint getJoystickCode(SDL_Event event)
	{
		if (event.type == SDL_JOYAXISMOTION)
		{
			if ((prejcs == Jcs.FREE && jcs == Jcs.UP)
				|| (prejcs == Jcs.UP && jcs == Jcs.FREE)) return 0;
			else if ((prejcs == Jcs.FREE && jcs == Jcs.DOWN)
				|| (prejcs == Jcs.DOWN && jcs == Jcs.FREE)) return 1;
			else if ((prejcs == Jcs.FREE && jcs == Jcs.RIGHT)
				|| (prejcs == Jcs.RIGHT && jcs == Jcs.FREE)) return 2;
			else if ((prejcs == Jcs.FREE && jcs == Jcs.LEFT)
				|| (prejcs == Jcs.LEFT && jcs == Jcs.FREE)) return 3;
			return uint.max;
		}
		else if (event.type == SDL_JOYBUTTONDOWN
			|| event.type == SDL_JOYBUTTONUP)
		{
			return event.jbutton.button + 4;
		}
		error(__LINE__);
		return 0;
	}
	
	private void delegate(uint input) _inputHandler;
	void inputHandler(void delegate(uint input) dg) { _inputHandler = dg; }
	
	private uint INPUT_BUFFER_MAX = 256;
	private uint[] inputBuffer;
	
	int opApply(int delegate(inout uint) dg)
	{
		while (inputBuffer.length != 0)
		{
			int result = dg(inputBuffer[0]);
			inputBuffer = inputBuffer[1..$];
			if (result) return result;
		}
		return 0;
	}
}

private struct InputEvent
{
	enum DeviceType { MOUSE, KEYBOARD, JOYSTICK }
	DeviceType deviceType;
	bit isMouse() { return deviceType == DeviceType.MOUSE ? true : false; }
	bit isKeyboard() { return deviceType == DeviceType.KEYBOARD ? true : false; }
	bit isJoystick() { return deviceType == DeviceType.JOYSTICK ? true : false; }
	bit isOn;
	bit isOff() { return !isOn ? true : false; }
	uint code;
	bit isMouseLeft() { return isMouse && code == 1 ? true : false; }
	bit isMouseRight() { return isMouse && code == 3 ? true : false; }
}

private class InputEventTranslator : Database
{
	private import std.conv;
	
	unittest
	{
		static assert(97 == 'a');
		const char[] rj = "
			result: 0
			device: mouse
			state: on
			code: 1
			%% マウス左 on
			
			result: 1
			device: keyboard
			state: off
			code: 97
			%% 'a' off
			
			result: 2
			device: joystick
			state: on
			code: 2
			%% ジョイスティック右 on
		";
		
		InputEventTranslator tra = new InputEventTranslator(rj);
		InputEvent e;
		e.deviceType = InputEvent.DeviceType.MOUSE;
		e.isOn = true;
		e.code = 1;
		assert(0 == tra.execute(e));
		
		e.deviceType = InputEvent.DeviceType.KEYBOARD;
		e.isOn = false;
		e.code = 'a';
		assert(1 == tra.execute(e));
		
		e.deviceType = InputEvent.DeviceType.JOYSTICK;
		e.isOn = true;
		e.code = 2;
		assert(2 == tra.execute(e));
		
		e.code = 99;
		assert(uint.max == tra.execute(e));
	}
	
	private final uint[InputEvent] eventMap;
	
	this(char[] script)
	{
		super(script, 0);
		foreach (Record record; this)
		{
			InputEvent key;
			key.deviceType = record["device"] == "mouse" ? InputEvent.DeviceType.MOUSE
				: record["device"] == "keyboard" ? InputEvent.DeviceType.KEYBOARD
				: InputEvent.DeviceType.JOYSTICK;
			key.isOn = record["state"] == "on" ? true : false;
			key.code = toUint(record["code"]);
			eventMap[key] = toUint(record["result"]);
		}
	}
	
	uint execute(InputEvent event)
	{
		if (!(event in eventMap)) return uint.max;
		return eventMap[event];
	}
}
