﻿module hdk.zipex;
private import std.zip;
private import std.string;
private import std.utf;
private import std.file;

///
class ZipArchiveEx : ZipArchive
{
	this() { super(); } ///
	this(void[] data) { super(data); } ///
	
	void add(char[] name, void[] data) ///
	{
		addMember(createArchiveMember(name, data));
	}
	
	private static ArchiveMember createArchiveMember(char[] name, void[] data) ///
	{
		ArchiveMember result = new ArchiveMember();
		result.name = name;
		result.expandedData = cast(ubyte[])data;
		result.compressionMethod = 8;
		return result;
	}
	
	
	unittest
	{
		const dchar[] TEXTURE = `test9data.zip\texture`;
		assert(isZip(TEXTURE));
		dchar[][] zt = TEXTURE.splitLastSlash();
		assert(2 == zt.length);
		assert(`test9data.zip` == zt[0]);
		assert("texture" == zt[1]);
		
		assert(isZip(`test9data.773\texture`));
	}
	
	static bit isZip(dchar[] fileName) ///
	{
		return -1 != toUTF8(fileName).ifind(`.zip\`)
			|| -1 != toUTF8(fileName).ifind(`.773\`) ? true : false;
	}
	
	unittest
	{
		dchar[] a = `てすと.773\vertex_buffer[0][0]`;
		assert(2 == splitLastSlash(a).length);
		assert(`てすと.773` == splitLastSlash(a)[0]);
		assert(`vertex_buffer[0][0]` == splitLastSlash(a)[1]);
	}
	
	static dchar[][] splitLastSlash(dchar[] a) ///
	{
		dchar[][] result;
		
		char[] utf8 = toUTF8(a);
		int s = utf8.rfind(`\`);
		
		result ~= toUTF32(utf8[0..s]);
		result ~= toUTF32(utf8[s + 1..$]);
		
		return result;
	}
	
	unittest
	{
		const uint WIDTH = 256, HEIGHT = 256;
		uint[] textureData = new uint[WIDTH * HEIGHT];
		for (int i = 0; i < textureData.length; i++)
		{
			int x = i % WIDTH;
			int y = i / WIDTH;
			textureData[i] = 0xff000000 | (y << 8) | (x);
		}
		
		ZipArchiveEx encoder = new ZipArchiveEx();
		encoder.addTexture("texture", textureData, WIDTH, HEIGHT);
		
		void[] zip = encoder.build();
		
		uint width, height;
		ZipArchiveEx decoder = new ZipArchiveEx(zip);
		uint[] result = decoder.extractTexture("texture", width, height);
		
		assert(WIDTH == width);
		assert(HEIGHT == height);
		assert(textureData == result);
	}
	
	void addTexture(dchar[] name, uint[] data, uint width, uint height) ///
	{
		uint[] a;
		a ~= width;
		a ~= height;
		a ~= data;
		add(toUTF8(name), a);
	}
	
	uint[] extractTexture(dchar[] name, out uint width, out uint height) ///
	{
		uint[] result = cast(uint[])expand(directory[toUTF8(name)]);
		width = result[0];
		height = result[1];
		return result[2..$].dup;
	}
	
	
	unittest
	{
		const float[] a = [ 1.5, 2.5, 3.5 ];
		uint[] b = cast(uint[])a;
		assert(a == cast(float[])b);
	}
	
	void addVertexBuffer(dchar[] name, float[] data, uint sizeOfVertex, uint vertexLength) ///
	{
		addTexture(name, cast(uint[])data, sizeOfVertex, vertexLength);
	}
	
	float[] extractVertexBuffer(dchar[] name, out uint sizeOfVertex, out uint vertexLength) ///
	{
		return cast(float[])extractTexture(name, sizeOfVertex, vertexLength);
	}
	
	
	bit has(char[] name) ///
	{
		foreach (ArchiveMember a; directory)
		{
			if (a.name == name) return true;
		}
		return false;
	}
	
	void[] get(char[] name) ///
	{
		return expand(directory[name]);
	}
	
	unittest
	{
		ZipArchiveEx encoder = new ZipArchiveEx();
		encoder.add("hoge", "hogehoge");
		ZipArchiveEx decoder = new ZipArchiveEx(encoder.build());
		assert(decoder.has("hoge"));
		assert(!decoder.has("foo"));
		assert("hogehoge" == decoder.get("hoge"));
	}
}

///
class BufferedZipFile
{
	private static ZipArchiveEx[dchar[]] map;
	
	private static ZipArchiveEx getZipArchiveEx(dchar[] zipFileName)
	{
		if (!(zipFileName in map))
		{
			map[zipFileName] = new ZipArchiveEx(read(toUTF8(zipFileName)));
		}
		return map[zipFileName];
	}
	
	static void clear()
	{
		map = null;
		assert(0 == map.length);
	}
	
	/**
	 * Params:
	 *  fileAndEntryName = `hoge.zip\entry`のように使う
	 */
	static uint[] readTexture(dchar[] fileAndEntryName, out uint width, out uint height)
	{
		dchar[] zipFileName, entryName;
		getName(fileAndEntryName, zipFileName, entryName);
		return getZipArchiveEx(zipFileName).extractTexture(entryName, width, height);
	}
	
	/// readTextureと同じように使う
	static float[] readVertexBuffer(dchar[] fileAndEntryName, out uint sizeOfVertex, out uint vertexLength)
	{
		dchar[] zipFileName, entryName;
		getName(fileAndEntryName, zipFileName, entryName);
		return getZipArchiveEx(zipFileName).extractVertexBuffer(entryName, sizeOfVertex, vertexLength);
	}
	
	static bit isZipEntry(dchar[] fileAndEntryName) ///
	{
		dchar[] zipFileName, entryName;
		getName(fileAndEntryName, zipFileName, entryName);
		char[] entryName2 = toUTF8(entryName);
		return entryName2 in getZipArchiveEx(zipFileName).directory ? true : false;
	}
	
	private static void getName(dchar[] fileAndEntryName, out dchar[] file, out dchar[] entry)
	{
		file = ZipArchiveEx.splitLastSlash(fileAndEntryName)[0];
		entry = ZipArchiveEx.splitLastSlash(fileAndEntryName)[1];
	}
	
	static uint getEntrySize(dchar[] fileAndEntryName) ///
	{
		dchar[] zipFileName, entryName;
		getName(fileAndEntryName, zipFileName, entryName);
		char[] entryName2 = toUTF8(entryName);
		return getZipArchiveEx(zipFileName).directory[entryName2].expandedSize;
	}
}
